import threading
from datetime import datetime

import numpy
import pause
import sortedcontainers
import sounddevice

from source import packets, managers
from source.utils.audio.audio import sample_width_to_type


class AudioManager:
    def __init__(self, manager: "managers.Manager"):
        # buffer containing the set of audio chunk to play. Sort them by their time to play
        self.buffer = sortedcontainers.SortedList(key=lambda audio: audio.time)

        # thread support
        self.lock = threading.Lock()
        self.new_audio_event = threading.Event()  # event triggered when a new audio have been added

    def add_audio(self, audio: packets.AudioPacket) -> None:
        """
        Add a new audio chunk to play
        :param audio: the audio chunk to play
        """

        with self.lock:
            # add the audio packet to the buffer
            self.buffer.add(audio)
            # trigger the new audio event
            self.new_audio_event.set()

    def handle(self) -> None:
        """
        Play the audio chunk in the buffer at the given time
        """

        # wait for a new audio packet
        self.new_audio_event.wait()

        # get the most recent audio packet to play
        audio: packets.AudioPacket = self.buffer.pop(0)

        # if the audio should have been played before, skip it
        if audio.time < datetime.now():
            return

        # create a numpy array for our sample
        sample = numpy.frombuffer(audio.data, dtype=sample_width_to_type(numpy.int16))
        # reshape it to have a sub-array for each channels
        sample = sample.reshape((-1, audio.channels))
        # normalize the sample to be between -1 and 1
        sample = sample / (2 ** (audio.sample_width * 8 - 1))

        # wait for the audio given time
        pause.until(audio.time)

        # play the audio
        sounddevice.play(sample, audio.sample_rate)

    def loop(self) -> None:
        """
        Handle forever
        """

        while True:
            self.handle()
